﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectのランダムドットパターンで赤外線脈波を検出』
//【名称】Kinect_HeartRateMonitor　　【Version】SDK Ver.1.6対応版
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章  Copyright by ☆Tomoaki Ueda☆
//【作成日】2013年01月14日
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Media;                   //【SoundPlayer】
using System.Diagnostics;             //【Debug】
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Imaging;         //【BitmapData】
using System.Runtime.InteropServices; //【Marshal.Copy】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択
using NTL.KinectLib;                  //【Neo-Tech-Lab.comのKinectライブラリ】

namespace Kinect_VitalMonitor5
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        private KinectSensor KS_KinectSensor;  //【KINECTセンサ】★利用可能なKinectセンサ
        private int KS_HeadAngle = 15;         //【Kinectヘッドの仰角を設定】設定範囲は±27°
        private int KS_Distance = 6000;        //【0～6000mmまでカラールックアップテーブルで表示】
        private Boolean flagReverse = false;   //【反転表示フラグ】
        private Boolean flagShowDepth = false; //【デプスイメージ表示フラグ】
        private ImageCamera NTLCamera = new ImageCamera(); //【NTL.KinectLibの汎用カメラ】
        private ColorBar IrBar = new ColorBar();
        private ColorBar BrBar = new ColorBar();
        private Bitmap IrScale, BrScale;
        private Rectangle IrRect = new Rectangle(290, 210, 60, 60); //【脈拍観測枠】
        private Rectangle BrRect = new Rectangle(220, 410, 50, 50); //【呼吸観測枠】
        private int IrThreshold = 10000; //【赤外線しきい値】
        private int IrCount = 0;         //【計測回数】
        private int BrCount = 0;         //【デプス計測回数】
        private int AccCount = 0;        //【加速度センサ用計測回数】
        private RectangleWaveFilter IrFilter = new RectangleWaveFilter();
        private NoiseReduction IrNRFilter = new NoiseReduction();
        private NoiseReduction BrNRFilter = new NoiseReduction();
        private LowPassFilter IrLPF = new LowPassFilter();
        private LowPassFilter BrLPF = new LowPassFilter();
        private FIFO IrFIFO = new FIFO();
        private FIFO BrFIFO = new FIFO();

        private const int nFIFO = 601;   //【波形表示の為のデータ格納配列サイズ】
        private const int nOffset = 300; //【オフセット除去処理サンプル数】
        private const int nWave = 24;    //【矩形波相関フィルタ処理サンプル数】★必ず4の倍数
        private const int nLPF = 12;     //【赤外線脈波のLPF処理サンプル数】
        private const int nBrLPF = 30;   //【呼吸ノイズ除去処理サンプル数】

        private double BrValue = 0.0;  //【単位mm】【呼吸値】
        private double BrOffset = 0.0; //【単位mm】【呼吸オフセット】
        private double IrValue = 0.0;  //【単位digit】【脈波値】
        private double IrOffset = 0.0; //【単位digit】【脈波オフセット】

        public Form1()
        {
            InitializeComponent();

            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            KS_KinectSensor = NTLCamera.DetectKinect(); //【利用可能なKINECTセンサを取得する】
            //【Kinect検出結果の表示】
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = imageList1.Images[1];
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else //【Kinect検出に成功した場合】
            {
                pictureBox2.Image = imageList1.Images[0];
            }

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            NTLCamera.Initialize(false,  //【RGBカメラ】 未使用
                                  true,  //【赤外線カメラ】使用 640x480Fps30
                                  true,  //【デプスカメラ】使用 640x480Fps30
                                 false,  //【骨格追尾】　未使用
                          KS_HeadAngle); //【Kinectの仰角設定】
            textBox1.Text = KS_HeadAngle.ToString(); //【Kinect設定仰角の表示】

            IrBar.CreateColorPalletForInfrared();                   //【赤外線強度16bitをカラー変換】
            BrBar.CreateColorPalletForDistance(KS_Distance);        //【0～6000mm】
            IrScale = IrBar.CreateColorScale(400, 10, false);       //【赤外線カメラ画像用カラースケール】400x10画素、非反転
            BrScale = BrBar.CreateColorScale(400, 10, false);       //【デプスカメラ画像用カラースケール】400x10画素、非反転
            pictureBox3.Image = IrScale;                            //【カラーバーを400×10画素で作成して表示する】
            pictureBox3.SizeMode = PictureBoxSizeMode.StretchImage; //【カラーバーをpictureBox3枠内に表示】

            KS_KinectSensor.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(ColorImageReady); //【KINECTデータ取得イベントとしてColorImageReadyを設定】
            KS_KinectSensor.DepthFrameReady += new EventHandler<DepthImageFrameReadyEventArgs>(DepthImageReady); //【DepthImageReadyをcallするよう設定】
            checkBox1.CheckedChanged += new EventHandler(checkBox1_CheckedChanged); //【左右反転】
            checkBox2.CheckedChanged += new EventHandler(checkBox2_CheckedChanged); //【赤外線プロジェクタの照射のOn/Off】
            checkBox3.CheckedChanged += new EventHandler(checkBox3_CheckedChanged); //【赤外線カメラ/デプス表示】
        }

        //***************************************************************
        //【左右表示反転フラグの処理】
        //***************************************************************
        void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            flagReverse = !flagReverse;
        }

        //***************************************************************
        //【赤外線プロジェクタの照射のOn/Offを切り替える】
        //***************************************************************
        private void checkBox2_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox2.Checked) { KS_KinectSensor.ForceInfraredEmitterOff = true; }
            else { KS_KinectSensor.ForceInfraredEmitterOff = false; }
        }

        //***************************************************************
        //【デプス表示フラグの処理】
        //***************************************************************
        void checkBox3_CheckedChanged(object sender, EventArgs e)
        {
            flagShowDepth = !flagShowDepth;
            if (flagShowDepth)
            {
                pictureBox3.Image = BrScale;
            }
            else
            {
                pictureBox3.Image = IrScale;
            }
        }

        //***************************************************************
        //【計測リセット】
        //***************************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            IrCount = 0;
            BrCount = 0;
            AccCount = 0;
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            int angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_HeadAngle = angle;
                NTLCamera.SetElevationAngle(angle);
            }
        }

        //***************************************************************
        //【Kinectの終了処理】
        //***************************************************************
        private void button3_Click(object sender, EventArgs e)
        {
            NTLCamera.Finish();
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
            this.Close();
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        //★Kinect for Windows SDK Ver.1.6からデプス演算アルゴリズムが変更になり、Depthは最大10ｍまでとなりました。
        //　CopyDepthImagePixelDataToメソッドで取得します。但し、頻繁にチューニングが行われ、ノイズが発生します。
        //　取得形式はDepthImagePixelの配列です。DepthプロパティーやPlayerIndexプロパティーを使います。
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (DepthImageFrame f = e.OpenDepthImageFrame())
            {
                if (f != null)
                {
                    f.CopyDepthImagePixelDataTo(NTLCamera.DepthImage); //【Depth情報を配列DepthImageにコピー】
                    if (BrCount == 0)
                    {
                        BrNRFilter.Initialize(nBrLPF);    //【最小二乗法でノイズ除去】【1秒分】
                        BrLPF.Initialize(nBrLPF);         //【LPFでノイズ除去】【1秒分】
                        BrFIFO.Initialize(nFIFO,nOffset); //【波形表示20秒分,オフセット処理10秒分】
                    }
                    double Data = NTLCamera.GetRectangleDepthAverage(BrRect, f.MinDepth, f.MaxDepth); //【枠内デプス値加算平均】
                    BrNRFilter.Estimation((float)Data); //【最小二乗法でノイズ除去】
                    Data = BrNRFilter.C;                //　：
                //  Data = BrLPF.LPF(Data);             //【LPFでノイズ除去】★上のフィルタで十分だったので未使用
                    BrFIFO.LoadDataToFIFO(Data);        //【波形表示用FIFOに格納】
                    BrValue = Data;
                    BrOffset = BrFIFO.OffsetLevel;
                    BrCount++;
                }
                else
                {
                    //【DepthFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }
        
        //***************************************************************
        //【ColorImageの取得イベント】
        //***************************************************************
        // Kinectからカラー画像１フレーム分が到着する度に、この関数が呼ばれます。
        // Windowsパソコンが最も高速にBitmapデータを表示できるデータ形式は1画素４バイトの形式なのでXRGB形式を使っています。
        // 但し、追加のCG描画で半透明表現を使うときはARGB形式を使います。
        unsafe private void ColorImageReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (ColorImageFrame f = e.OpenColorImageFrame())
            {
                if (f != null)
                {
                    //【赤外線カメラ画像の取得と表示】
                    f.CopyPixelDataTo(NTLCamera.ImageData); //【KINECTのColor画像を配列KS_InfraredImageDataにコピー】
                    if (flagShowDepth)
                    {
                        pictureBox1.Image = NTLCamera.ConvertToDepthImage(flagReverse);    //【pictureBox1に画像表示】
                    }
                    else
                    {
                        pictureBox1.Image = NTLCamera.ConvertToInfraredImage(flagReverse); //【pictureBox1に画像表示】
                    }

                    if (IrCount == 0) //【初回はFIFOを初期化する】
                    {
                        IrFilter.Initialize(nWave);        //【矩形波相関フィルタの初期化】★4の倍数を指定すること
                        IrNRFilter.Initialize(nWave);      //【最小２乗法でノイズ除去】
                        IrLPF.Initialize(nLPF);            //【LPFの初期化】
                        IrFIFO.Initialize(nFIFO, nOffset); //【波形表示20秒分,オフセット処理10秒分】
                    }
                    double Data = NTLCamera.GetInfraredIntensity(IrRect, IrThreshold); //【距離測定】指定領域内の加算平均デプス値を求める
                    Data = IrFilter.CrossCorrelationFilter(Data); //【矩形波相関フィルタ処理】
                    IrNRFilter.Estimation((float)Data);           //【高性能ノイズ除去フィルタ】
                    Data = IrNRFilter.C;                          //【高性能ノイズ除去フィルタ】
                //  Data = IrLPF.LPF(Data);                       //【LPF】★上のフィルタで十分だったので未使用
                    IrFIFO.LoadDataToFIFO(Data);                  //【波形表示用FIFOに格納】
                    IrValue = Data;
                    IrOffset = BrFIFO.OffsetLevel;

                    //【波形表示処理】
                    Graphics g = Graphics.FromImage(pictureBox1.Image); //【pixtureBox1内に測定値を重畳表示】
                    DisplayGrid(g, Color.HotPink);                      //【グリッドの表示】
                    DisplayRectangleFrame(g, IrRect, Color.SkyBlue);    //【赤外線脈波枠の描画】
                    DisplayRectangleFrame(g, BrRect, Color.LightGreen); //【呼吸測定枠の描画】
                    DisplayHeartRateSignal(g, Color.Red);
                    if(BrCount>0)
                    {
                        DisplayBreathingSignal(g, Color.LightGreen);
                    }
                    g.Dispose();
                    DisplayLabels(); //【数値表示】6フレームに1回表示
                    IrCount++;       //【測定番号を+1する】
                }
                else
                {
                    //【ImageFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【指定フレーム数毎に最近点座標とKionix３軸直交加速度センサの測定値を表示】
        //***************************************************************
        private void DisplayLabels()
        {
            AccCount++;        //【表示用カウンタを+1】
            if (AccCount == 6) //【30FPSなら0.2秒に1回表示】
            {
                AccCount = 0;  //【表示用カウンタをクリア】

                //【バイタル測定値】
                label3.Text = "呼吸測定値=" + BrValue.ToString() + "[mm]";
                label4.Text = "呼吸ｵﾌｾｯﾄ =" + BrOffset.ToString() + "[mm]";
                label5.Text = "脈波測定値=" + IrValue.ToString() + "[digits]";
                label6.Text = "脈波ｵﾌｾｯﾄ =" + IrOffset.ToString() + "[digits]";

                //【内蔵加速度センサ】
                double AccX = NTLCamera.AccelerometerCurrentValue.X;
                double AccY = NTLCamera.AccelerometerCurrentValue.Y;
                double AccZ = NTLCamera.AccelerometerCurrentValue.Z;
                double AccABS = NTLCamera.AccelerometerCurrentValue.W;
                label9.Text  = "x=" + AccX.ToString() + "[g]";
                label10.Text = "y=" + AccY.ToString() + "[g]";
                label11.Text = "z=" + AccZ.ToString() + "[g]";
                label12.Text = "|Acc|=" + AccABS.ToString() + "[g]";
            }
        }

        //***************************************************************
        //【DisplayGrid】グリッドを表示する
        //***************************************************************
        private void DisplayGrid(Graphics g, Color lc)
        {
            //【縦は10分割、デプスは10mmppに相当、赤外線脈波は100degit】
            int i;
            for (i = 0; i < 640; i += 60) //【時間軸は60で1秒】
            {
                g.DrawLine(new Pen(lc, 3.0f), i, 0, i, 480);       //【画面中央縦線】
            }
            g.DrawLine(new Pen(lc, 3.0f), 639, 0, 639, 480);       //【画面中央縦線】
            for (i = 0; i < 480; i += 48)
            {
                g.DrawLine(new Pen(lc, 3.0f), 0, i, 640, i);       //【画面中央横線】
            }
        }

        //***************************************************************
        //【DisplayRectangleFrame】矩形観測領域の枠を表示する
        //***************************************************************
        private void DisplayRectangleFrame(Graphics g, Rectangle rect, Color lc)
        {
            if (flagReverse) //【左右反転している場合の対処】
            {
                Rectangle rrect = new Rectangle(640 - rect.Location.X - rect.Size.Width,
                                                rect.Location.Y,
                                                rect.Size.Width,
                                                rect.Size.Height);
                g.DrawRectangle(new Pen(lc, 3.0f), rrect); //【観測領域】
            }
            else
            {
                g.DrawRectangle(new Pen(lc, 3.0f), rect); //【観測領域】
            }
        }

        //***************************************************************
        //【DisplayBreathingSignal】FIFOクラスを使って波形を表示する
        //***************************************************************
        private void DisplayBreathingSignal(Graphics g, Color lc)
        {
            int i, j;
            double Dy;
            const int FrameHeight = 480;
            Dy = FrameHeight / 10.0; //【デプス呼吸は10mm_pp】
            for (i = 0, j = BrFIFO.Pointer; i < nFIFO; i++)
            {
                BrFIFO.Points[i].X = (int)i; //0～600
                BrFIFO.Points[i].Y = (int)((BrFIFO.dFIFO[j] - BrFIFO.OffsetLevel) * Dy) + 240; //【中央値をOffsetLevelとする】
                j++; if (j == nFIFO) { j = 0; } //【FIFO管理ポインタを更新する】
            }
            g.DrawLines(new Pen(lc, 5.0f), BrFIFO.Points);
        }

        //***************************************************************
        //【DisplayHeartRateSignal】FIFOクラスを使って波形を表示する
        //***************************************************************
        private void DisplayHeartRateSignal(Graphics g, Color lc)
        {
            int i, j;
            double Dy;
            const int FrameHeight = 480;
            Dy = FrameHeight / 500.0; //【赤外線脈波は500digits_pp】
            for (i = 0, j = IrFIFO.Pointer; i < nFIFO; i++)
            {
                IrFIFO.Points[i].X = (int)i; //0～600
                IrFIFO.Points[i].Y = (int)((IrFIFO.dFIFO[j] - IrFIFO.OffsetLevel) * Dy) + 240; //【中央値をOffsetLevelとする】
                j++; if (j == nFIFO) { j = 0; } //【FIFO管理ポインタを更新する】
            }
            g.DrawLines(new Pen(lc, 5.0f), IrFIFO.Points);
        }

    }
}
