﻿//*************************************************************************
//【サンプルプログラム名称】『NTL.KinectLib.dll』
//【目的】Kinectアプリケーション試作用ライブラリ
//【名称】Kinect_HeartRateMonitor
//【Kinect for Windows SDK Version】1.7
//【NTL.KinectLib DLL Version】0.63
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章  Copyright by ☆Tomoaki Ueda☆
//【作成日】2013年03月20日
//*************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Imaging;         //【BitmapData】
using System.Runtime.InteropServices; //【Marshal.Copy】
using Microsoft.Kinect;


namespace NTL.KinectLib
{
    //*************************************************************************
    //*************************************************************************
    //【CLASS】『ImageCamera』【赤外線カメラ画像をキャプチャー】
    //*************************************************************************
    //*************************************************************************
    public class ImageCamera
    {
        //【Kinect】
        public KinectSensor KS;                  //【KinectSensor】
        private Boolean sts = false;             //【KinectSensorの接続状態】
        private Bitmap stsImage;                 //【KinectSensorの接続状態を示すbitmapイメージ】
        private Boolean flagColorImageStream;    //【フラグ】RGBカメラ
        private Boolean flagInfraredImageStream; //【フラグ】赤外線カメラ
        private Boolean flagDepthImageStream;    //【フラグ】デプスカメラ
        private Boolean flagSkeletonStream;      //【フラグ】骨格追尾
        private Boolean flagPlayer = false;      //【フラグ】プレーヤー検出
        private Vector4 Accelerometer;           //【Kinect内蔵の直交３軸加速度センサの測定値】
        public byte[] ImageData;             //【赤外線カメラ】赤外線カメラ画像16bit/画素,GrayScale表示形式
        public byte[] ImageDataXRGB;         //【RGBカメラ】32bit/画素,X8R8G8B8表示形式
        public CoordinateMapper map;
        public ColorImagePoint[] ColorPoint; //【マップ】【デプスイメージ】⇒【RGB画像】
        public DepthImagePixel[] DepthImage; //【デプスカメラ】.Depth 16bit
        public byte[] DepthImageXRGB;        //【擬似カラーデプス】32bit/画素,X8R8G8B8表示形式
        private int FrameWidth;              //【KINECTのColorFrameの横方向画素数】
        private int FrameHeight;             //【KINECTのColorFrameの縦方向画素数】
        private int DepthWidth;              //【KINECTのDepthFrameの横方向画素数】
        private int DepthHeight;             //【KINECTのDepthFrameの縦方向画素数】
        private Bitmap bar1;                 //【赤外線輝度カラーバーのBitmapオブジェクト】
        private Bitmap bar2;                 //【デプススケールバーのBitmapオブジェクト】
        private ColorBar InfraredColorBar = new ColorBar(); //【赤外線強度をカラーに変換する部品】
        private ColorBar DepthColorBar = new ColorBar();    //【デプス値をカラーに変換する部品】


        //【RGB/赤外線カメラ】
        private int AddCount = 0;            //【赤外線カメラ画像の加算回数】
        public double[] IrPattern;           //【ProjectionPattern】64回加算平均
        public double[] ConvPattern;         //【Pre-ConvolutedPattern】
        public double[] ConvFrame;
        private Bitmap bmp1;                 //【赤外線輝度表示用Bitmapオブジェクト】
        private BitmapData bmpdata1;         //【ColorImage用BitmapData】
        private Rectangle rect1;             //【ColorImage表示枠】
        private Bitmap bmp2;                 //【デプス表示用Bitmapオブジェクト】
        private BitmapData bmpdata2;         //【ColorImage用BitmapData】
        private Rectangle rect2;             //【ColorImage表示枠】
        private Bitmap bmp3;                 //【デプス表示用Bitmapオブジェクト】
        private BitmapData bmpdata3;         //【ColorImage用BitmapData】
        private Rectangle rect3;             //【ColorImage表示枠】
        private Bitmap bmp4;                 //【デプス表示用Bitmapオブジェクト】
        private BitmapData bmpdata4;         //【ColorImage用BitmapData】
        private Rectangle rect4;             //【ColorImage表示枠】
        private double[] Hosei = new double[65536];

        //【未公開部分】
        private const int nWidth = 640;
        private const int nSpan = 16;
        private const double nSpan1 = 1.0 / (double)nSpan;
        private const double nSpan2 = 1.0 / (double)(nSpan * nSpan);
        private const int nSpanOfs = nSpan >> 1;
        private double[] FIFOx = new double[nSpan];           //【1次元水平方向FIFO、サイズはNTL_nSpan】
        private double[] FIFOxy = new double[nWidth * nSpan]; //【2次元矩形FIFO、サイズはNTL_nSpan*KS_ColorFrameWidth】
        private double[] SumXY = new double[nWidth];          //【矩形領域移動平均、サイズはKS_ColorFrameWidth】

        //*******************************************************
        //【DetectKinect】Kinectを検出する
        //*******************************************************
        public KinectSensor DetectKinect()
        {
            string filename;
            //【Kinectを検出する】
            KS = (from sensorToCheck
                  in KinectSensor.KinectSensors
                  where sensorToCheck.Status == KinectStatus.Connected
                  select sensorToCheck).FirstOrDefault();
            if (KS != null)
            {
                sts = true;
                filename = "Image/KinectOn.png";
            }
            else
            {
                sts = false;
                filename = "Image/KinectOff.png";

            }
            stsImage = (Bitmap)Image.FromFile(filename);
            map = new CoordinateMapper(KS);
            return KS;
        }

        //*******************************************************
        //【Status】Kinectの接続状態を取得する
        //*******************************************************
        public Boolean Status
        {
            get
            {
                return sts;
            }
        }


        //*******************************************************
        //【StatusImage】Kinectの接続状態を示すイメージを取得する
        //*******************************************************
        public Bitmap StatusImage
        {
            get
            {
                return stsImage;
            }
        }


        //*******************************************************
        //【Initialize】Kinectを赤外線カメラ画像のキャプチャが行えるように初期化する
        //*******************************************************
        public void Initialize(Boolean fColorStream,    //【ColorStream許可】
                               Boolean fInfraredStream, //【InfraredStream許可】ColorStream未使用時のみ可
                               Boolean fDepthStream,    //【DepthStream許可】
                               Boolean fSkeletonStream, //【SkeletonStream許可】
                               int ElevationAngle)      //【仰角：設定角度】±27°
        {
            flagColorImageStream = fColorStream;
            flagInfraredImageStream = fInfraredStream;
            flagDepthImageStream = fDepthStream;
            flagSkeletonStream = fSkeletonStream;

            //【Stream設定後Kinect動作開始】
            if (fColorStream)        //【ColorStream許可】
            {
                KS.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30);
            }
            else
            {
                if (fInfraredStream) //【InfraredStream許可】
                {
                    KS.ColorStream.Enable(ColorImageFormat.InfraredResolution640x480Fps30);
                }
            }
            if (fDepthStream)        //【DepthStream許可】
            {
                KS.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);
            }
            if (fSkeletonStream)     //【SkeletonStream許可】
            {
                KS.SkeletonStream.Enable();
            }
            KS.Start();              //【KINECT動作開始】

            if (fColorStream || fInfraredStream) //【カラー/赤外線カメラ関連】
            {
                FrameWidth = KS.ColorStream.FrameWidth;               //【カラー/赤外線カメラ画像の横方向画素数】
                FrameHeight = KS.ColorStream.FrameHeight;             //【カラー/赤外線カメラ画像の縦方向画素数】
                rect1 = new Rectangle(0, 0, FrameWidth, FrameHeight); //【ColorImage用枠】
                bmp1 = new Bitmap(FrameWidth, FrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【ColorImage表示用】
                ColorPoint = new ColorImagePoint[FrameWidth * FrameHeight]; //【マップ】
            }
            if (fColorStream) //【カラーカメラ関連】
            {
                ImageDataXRGB = new byte[KS.ColorStream.FramePixelDataLength]; //【カラー表示用フレームメモリを確保】
            }
            if (fInfraredStream) //【赤外線カメラ関連】
            {
                ImageDataXRGB = new byte[2 * KS.ColorStream.FramePixelDataLength]; //【カラー表示用フレームメモリを確保】
                ImageData = new byte[KS.ColorStream.FramePixelDataLength]; //【赤外線カメラ画像の生データ】
                IrPattern = new double[FrameWidth * FrameHeight];          //【投影しているランダムドットパターン】
                ConvPattern = new double[FrameWidth * FrameHeight];        //【Pre-Convoluted Code Pattern】
                ConvFrame = new double[FrameWidth * FrameHeight];          //
                InfraredColorBar.CreateColorPalletForInfrared();           //【赤外線強度16bitをカラー変換するパレットを作成】
                bar1 = InfraredColorBar.CreateColorScale(512, 10, false);  //【カラーバーのbitmapを512×10画素で作成】
                rect3 = new Rectangle(0, 0, FrameWidth, FrameHeight);              //【ColorImage用枠】
                bmp3 = new Bitmap(FrameWidth, FrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【ColorImage表示用】
                rect4 = new Rectangle(0, 0, FrameWidth, FrameHeight);              //【ColorImage用枠】
                bmp4 = new Bitmap(FrameWidth, FrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【ColorImage表示用】
                CalHosei();
            }
            if (fDepthStream) //【デプスカメラ関連】
            {
                DepthWidth = KS.DepthStream.FrameWidth;                     //【デプスカメラ画像の横方向画素数】
                DepthHeight = KS.DepthStream.FrameHeight;                   //【デプスカメラ画像の縦方向画素数】
                rect2 = new Rectangle(0, 0, DepthWidth, DepthHeight);       //【DepthImage用枠】
                bmp2 = new Bitmap(DepthWidth, DepthHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【ColorImage表示用】
                DepthImage = new DepthImagePixel[DepthWidth * DepthHeight]; //【デプスイメージピクセル】
                DepthImageXRGB = new byte[4 * DepthWidth * DepthHeight];    //【デプス表示用フレームメモリを確保】
                DepthColorBar.CreateColorPalletForDistance();               //【デプス値をカラー変換するパレットを作成】
                bar2 = DepthColorBar.CreateColorScale(512, 10, false);      //【カラーバーのbitmapを512×10画素で作成】
                KS.DepthStream.Range = DepthRange.Near;                     //【デプス測定モードをNearに設定】
            }
            KS.CoordinateMapper.MapDepthFrameToColorFrame(
                DepthImageFormat.Resolution640x480Fps30,
                DepthImage,
                ColorImageFormat.RgbResolution640x480Fps30,
                ColorPoint);
            SetElevationAngle(ElevationAngle); //【Kinectの仰角を変更する】
        }

        //*******************************************************
        //【Finish】Kinectの利用を終了する
        //*******************************************************
        public void Finish()
        {
            if (KS != null)
            {
                KS.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
        }

        //*******************************************************
        //【AccelerometerCurrentValue】Kinect内蔵直交３軸加速度センサの測定値を取得する
        //*******************************************************
        public Vector4 AccelerometerCurrentValue
        {
            get
            {
                Accelerometer = KS.AccelerometerGetCurrentReading();
                float AccX = Accelerometer.X;
                float AccY = Accelerometer.Y;
                float AccZ = Accelerometer.Z;
                Accelerometer.W = (float)Math.Sqrt((double)(AccX * AccX + AccY * AccY + AccZ * AccZ));
                return Accelerometer;
            }
        }

        //*******************************************************
        //【ConvertToColorImage】カラーカメラ画像を取得する
        //*******************************************************
        public Bitmap TransferColorImage(Boolean fReverse)
        {
            bmpdata1 = bmp1.LockBits(rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
            Marshal.Copy(ImageDataXRGB, 0, bmpdata1.Scan0, ImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp1.UnlockBits(bmpdata1);                                           //【Lock解除】
            if (fReverse) { bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp1;
        }

        //*******************************************************
        //【ConvertToInfraredImage】赤外線カメラ画像(16bit/画素)をカラー変換する
        //*******************************************************
        public Bitmap ConvertToInfraredImage(Boolean fReverse)
        {
            ConvertInfraredToColor();
            bmpdata1 = bmp1.LockBits(rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
            Marshal.Copy(ImageDataXRGB, 0, bmpdata1.Scan0, ImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp1.UnlockBits(bmpdata1);                                           //【Lock解除】
            if (fReverse) { bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp1;
        }

        //*******************************************************
        //【ConvertInfraredToColor】16ビットのInfrared値をカラーバーに対応する色に変更
        //*******************************************************
        public void ConvertInfraredToColor()
        {
            int i, j, data;
            for (i = j = 0; i < ImageData.Length; i += 2)
            {
                data = ImageData[i + 1] * 256 + ImageData[i]; //【IR値16bitからパレット番号取得】
                data = (int)Hosei[data];
                ImageDataXRGB[j] = InfraredColorBar.B[data];
                ImageDataXRGB[j + 1] = InfraredColorBar.G[data];
                ImageDataXRGB[j + 2] = InfraredColorBar.R[data];
                ImageDataXRGB[j + 3] = InfraredColorBar.A[data];
                j += 4;
            }
        }

        //***************************************************************
        //【GetRectangleInfraredIntensity】
        //***************************************************************
        //　閾値NTL_Thresholdを超えた赤外線強度サンプルの領域内平均値を得る
        public double GetInfraredIntensity(Rectangle rect, int Threshold)
        {
            int x0 = rect.Location.X;
            int y0 = rect.Location.Y;
            int w = rect.Size.Width;
            int h = rect.Size.Height;
            double sum = 0.0f;
            int i, j, k, L, n = 0, v;
            for (j = y0; j < y0 + h; j++)
            {
                k = FrameWidth * j;
                for (i = x0; i < x0 + w; i++)
                {
                    L = 2 * (i + k);
                    v = (int)ImageData[L + 1] * 256 + (int)ImageData[L];
                    if (v > Threshold) { sum += v; n++; }
                }
            }
            if (n == 0) { n = 1; }
            sum /= (double)(n);
            return sum;
        }

        //*******************************************************
        //【ConvertToDepthImage】デプスカメラ画像(16bit/画素)をカラー変換する
        //*******************************************************
        public Bitmap ConvertToDepthImage(Boolean fReverse)
        {
            ConvertDepthToColor();
            bmpdata2 = bmp2.LockBits(rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
            Marshal.Copy(DepthImageXRGB, 0, bmpdata2.Scan0, DepthImageXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp2.UnlockBits(bmpdata2);                                              //【Lock解除】
            if (fReverse) { bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp2;
        }

        //*******************************************************
        //【ConvertToDepthImageWithAlpha】デプスカメラ画像(16bit/画素)をカラー変換(透明プレーン付)する
        //*******************************************************
        public Bitmap ConvertToDepthImageWithAlpha(Boolean fReverse)
        {
            ConvertDepthToColorWithAlpha();
            bmp2 = new Bitmap(DepthWidth, DepthHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【ColorImage表示用】
            bmpdata2 = bmp2.LockBits(rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【Lock】
            Marshal.Copy(DepthImageXRGB, 0, bmpdata2.Scan0, DepthImageXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp2.UnlockBits(bmpdata2);                                              //【Lock解除】
            if (fReverse) { bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp2;
        }

        public Bitmap ConvertToDepthImageWithAlpha(Boolean fReverse, byte Alpha)
        {
            ConvertDepthToColorWithAlpha(Alpha);
            bmp2 = new Bitmap(DepthWidth, DepthHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【ColorImage表示用】
            bmpdata2 = bmp2.LockBits(rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【Lock】
            Marshal.Copy(DepthImageXRGB, 0, bmpdata2.Scan0, DepthImageXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp2.UnlockBits(bmpdata2);                                              //【Lock解除】
            if (fReverse) { bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp2;
        }

        //*******************************************************
        //【TransferDepthImage】デプスカメラ画像(16bit/画素)をBitmapに変換する
        //*******************************************************
        public Bitmap TransferDepthImage(Boolean fReverse)
        {
            bmpdata2 = bmp2.LockBits(rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
            Marshal.Copy(DepthImageXRGB, 0, bmpdata2.Scan0, DepthImageXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp2.UnlockBits(bmpdata2);                                              //【Lock解除】
            if (fReverse) { bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp2;
        }

        //*******************************************************
        //【ConvertInfraredToColor】16ビットのInfrared値をカラーバーに対応する色に変更
        //*******************************************************
        public void ConvertDepthToColor()
        {
            int i, j, data;
            int L = DepthColorBar.A.Length - 1;
            for (i = j = 0; i < DepthImage.Length; i++)
            {
                data = DepthImage[i].Depth; //【IR値16bitからパレット番号取得】
                if (data < 0) { data = 0; }
                if (data > L) { data = L; }
                DepthImageXRGB[j] = DepthColorBar.B[data];
                DepthImageXRGB[j + 1] = DepthColorBar.G[data];
                DepthImageXRGB[j + 2] = DepthColorBar.R[data];
                DepthImageXRGB[j + 3] = DepthColorBar.A[data];
                j += 4;
            }
        }

        //*******************************************************
        //【ConvertInfraredToColorWithAlpha】16ビットのInfrared値をカラーバーに対応する色に変更
        //*******************************************************
        public void ConvertDepthToColorWithAlpha()
        {
            int i, j, data; //【Normalモード】800～4000mm【Nearモード】  400～3000mm
            int playerID;   //【下位3bitはplayerID】【未使用】
            int f;          //【PlayerLost検出用】
            int L = DepthColorBar.A.Length - 1;
            if (flagPlayer == false) //【player未検出時はDepthImage全体を表示する】
            {
                j = 0;
                for (i = 0; i < DepthImage.Length; i++)
                {
                    data = DepthImage[i].Depth;           //【深度情報】
                    playerID = DepthImage[i].PlayerIndex; //【short型playerID】
                    if (playerID > 0) { flagPlayer = true; }
                    if (data < 0) { data = 0; }
                    if (data > L) { data = L; }
                    DepthImageXRGB[j] = DepthColorBar.B[data];
                    DepthImageXRGB[j + 1] = DepthColorBar.G[data];
                    DepthImageXRGB[j + 2] = DepthColorBar.R[data];
                    DepthImageXRGB[j + 3] = DepthColorBar.A[data]; //【128程度にすると半透明になる。】
                    j += 4;
                }
            }
            else //【player検出に成功した場合にはPlayerのいる場所だけ表示して、他の部分は透明にする】
            {
                j = 0; f = 0;
                for (i = 0; i < DepthImage.Length; i++)
                {
                    data = DepthImage[i].Depth;           //【深度情報】
                    playerID = DepthImage[i].PlayerIndex; //【short型playerID】
                    if (playerID > 0)
                    {
                        f = 1;
                        if (data < 0) { data = 0; }
                        if (data > L) { data = L; }
                        DepthImageXRGB[j] = DepthColorBar.B[data];
                        DepthImageXRGB[j + 1] = DepthColorBar.G[data];
                        DepthImageXRGB[j + 2] = DepthColorBar.R[data];
                        DepthImageXRGB[j + 3] = DepthColorBar.A[data];
                    }
                    else
                    {   //【背景画像を透過させるために透明にする】
                        DepthImageXRGB[j] = 0;     //【青】B
                        DepthImageXRGB[j + 1] = 0; //【緑】G
                        DepthImageXRGB[j + 2] = 0; //【赤】R
                        DepthImageXRGB[j + 3] = 1; //【α】
                    }
                    j += 4;
                }
                if (f == 0) { flagPlayer = false; } //【Playerがいなくなったようなのでモードを通常のデプスイメージ表示に戻す】
            }
        }

        public void ConvertDepthToColorWithAlpha(byte Alpha)
        {
            int i, j, data; //【Normalモード】800～4000mm【Nearモード】  400～3000mm
            int playerID;   //【下位3bitはplayerID】【未使用】
            int f;          //【PlayerLost検出用】
            int L = DepthColorBar.A.Length - 1;
            if (flagPlayer == false) //【player未検出時はDepthImage全体を表示する】
            {
                j = 0;
                for (i = 0; i < DepthImage.Length; i++)
                {
                    data = DepthImage[i].Depth;           //【深度情報】
                    playerID = DepthImage[i].PlayerIndex; //【short型playerID】
                    if (playerID > 0) { flagPlayer = true; }
                    if (data < 0) { data = 0; }
                    if (data > L) { data = L; }
                    DepthImageXRGB[j] = DepthColorBar.B[data];
                    DepthImageXRGB[j + 1] = DepthColorBar.G[data];
                    DepthImageXRGB[j + 2] = DepthColorBar.R[data];
                    DepthImageXRGB[j + 3] = Alpha; //【128程度にすると半透明になる。】
                    j += 4;
                }
            }
            else //【player検出に成功した場合にはPlayerのいる場所だけ表示して、他の部分は透明にする】
            {
                j = 0; f = 0;
                int x, y, k;
                for (i = 0; i < DepthImage.Length; i++)
                {
                    data = DepthImage[i].Depth;           //【深度情報】
                    playerID = DepthImage[i].PlayerIndex; //【short型playerID】
                    x = ColorPoint[i].X; y = ColorPoint[i].Y;
                    j = 640 * y + x;
                    if (j < ColorPoint.Length)
                    {
                        j *= 4;
                        if (playerID > 0)
                        {
                            f = 1;
                            if (data < 0) { data = 0; }
                            if (data > L) { data = L; }
                            DepthImageXRGB[j] = DepthColorBar.B[data];
                            DepthImageXRGB[j + 1] = DepthColorBar.G[data];
                            DepthImageXRGB[j + 2] = DepthColorBar.R[data];
                            DepthImageXRGB[j + 3] = Alpha;
                        }
                        else
                        {   //【背景画像を透過させるために透明にする】
                            DepthImageXRGB[j] = 0;     //【青】B
                            DepthImageXRGB[j + 1] = 0; //【緑】G
                            DepthImageXRGB[j + 2] = 0; //【赤】R
                            DepthImageXRGB[j + 3] = 1; //【α】
                        }
                    }
                    //j += 4;

                }
                if (f == 0) { flagPlayer = false; } //【Playerがいなくなったようなのでモードを通常のデプスイメージ表示に戻す】
            }
        }

        //*******************************************************
        //【InfraredBarImage】赤外線輝度カラーバーのbitmapを取得
        //*******************************************************
        public Bitmap InfraredBarImage
        {
            get { return bar1; }
        }

        //***************************************************************
        //【GetRectangleDepthAverage】
        //　閾値KS_DepthFrame.MinDepthを超えたデプス値の指定領域(rect)内平均値を得る
        //　実際に測ると影や鏡面反射の影響で距離の取得できなかった点も混入するため
        //***************************************************************
        public double GetRectangleDepthAverage(Rectangle rect, int MinDepth, int MaxDepth)
        {
            double sum = 0.0;
            int i, j, k, n = 0, v;
            int x0 = rect.Location.X;
            int y0 = rect.Location.Y;
            int w = rect.Size.Width;
            int h = rect.Size.Height;
            for (j = y0; j < y0 + h; j++)
            {
                k = DepthWidth * j;
                for (i = x0; i < x0 + w; i++)
                {
                    v = (int)DepthImage[i + k].Depth;
                    if ((v > MinDepth) && (v < MaxDepth)) { sum += v; n++; }
                }
            }
            if (n != 0) { return sum / (double)n; } else { return 0.0; }
        }

        //*******************************************************
        //【InfraredBarImage】赤外線輝度カラーバーのbitmapを取得
        //*******************************************************
        public Bitmap DepthBarImage
        {
            get { return bar2; }
        }

        //*******************************************************
        //【SetElevationAngle】Kinectの仰角を変更する
        //*******************************************************
        public void SetElevationAngle(int angle)
        {
            if (angle < (-27)) { angle = (-27); }
            if (angle > 27) { angle = 27; }
            KS.ElevationAngle = angle;
        }

        //*******************************************************
        //【ElevationAngle】Kinectの仰角を取得する
        //*******************************************************
        public int ElevationAngle
        {
            get { return KS.ElevationAngle; }
        }



        //*******************************************************
        //【ClearPattern】赤外線カメラ画像のテンプレートを初期化する
        //*******************************************************
        public void ClearPattern()
        {
            AddCount = 0;
            for (int i = 0; i < FrameWidth * FrameHeight; i++)
            {
                IrPattern[i] = 0.0;
                ConvPattern[i] = 0.0;
            }
        }

        //*******************************************************
        //【DispIrPattern】赤外線カメラ画像を２値化してランダムドットパターンを抽出する
        //*******************************************************
        public void DispIrPattern()
        {
            int i;
            double X;
            double[] FIFO = new double[10];
            double sum = 0.0;
            int ptr = 0;
            for (i = 0; i < 10; i++) { FIFO[i] = 0.0; }
            for (i = 0; i < ImageData.Length; i += 2)
            {
                X = (double)(ImageData[i + 1] * 256 + ImageData[i]);
                sum = sum + X - FIFO[ptr];
                FIFO[ptr] = X;
                ptr++; if (ptr == 10) { ptr = 0; }
                if ((X * 10) > sum) { ImageData[i + 1] = ImageData[i] = 255; } else { ImageData[i + 1] = ImageData[i] = 0; }
            }
        }

        //*******************************************************
        //【AddPattern】赤外線カメラ画像をテンプレートIrPatternに加算する
        //*******************************************************
        public void AddPattern()
        {
            int j = 0;
            for (int i = 0; i < IrPattern.Length; i++)
            {
                IrPattern[i] += (double)(ImageData[j + 1] * 256 + ImageData[j]) - 32768.0; ;
                j += 2;
            }
            AddCount++; //【パターン加算回数を保存】
        }


        //*******************************************************
        //【AveragePattern】±1の加算平均テンプレートIrPatternを取得する
        //*******************************************************
        public Bitmap AveragePattern(Boolean fReverse)
        {
            double K1 = 1.0 / (double)AddCount;
            double K2 = 1.0 / 32678.0;
            int j = 0, data;
            double X;
            for (int i = 0; i < IrPattern.Length; i++)
            {
                X = IrPattern[i] * K1;
                IrPattern[i] = X * K2;
                data = (int)X + 32768;
                if (data > 65535) { data = 65535; }
                if (data < 0) { data = 0; }
                ImageDataXRGB[j] = InfraredColorBar.B[data];
                ImageDataXRGB[j + 1] = InfraredColorBar.G[data];
                ImageDataXRGB[j + 2] = InfraredColorBar.R[data];
                ImageDataXRGB[j + 3] = InfraredColorBar.A[data];
                j += 4;
            }
            bmpdata3 = bmp3.LockBits(rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
            Marshal.Copy(ImageDataXRGB, 0, bmpdata3.Scan0, ImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp3.UnlockBits(bmpdata3);                                              //【Lock解除】
            if (fReverse) { bmp3.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp3;
        }

        //*******************************************************
        //▲【Multiply1】赤外線画像とテンプレートを単純乗算して不一致65535、一致0となる画像を取得する(移動平均なし)
        //*******************************************************
        public void Multiply1()
        {
            int i, j, k;
            double d;
            for (i = j = 0; i < ImageData.Length; i += 2)
            {
                d = (double)(ImageData[i + 1] * 256 + ImageData[i]);
                k = (int)((32768.0 - d) * IrPattern[j]) + 32768;
                if (k < 0) { k = 0; }
                k = k & 0xffff;
                j++;
                ImageData[i + 1] = (byte)((k >> 8) & 0xff);
                ImageData[i] = (byte)(k & 0xff);
            }
        }

        //*******************************************************
        //【CreateConvPattern】±1の加算平均テンプレートIrPatternを取得する
        //*******************************************************
        public Bitmap CreateConvPattern(Boolean fReverse)
        {
            int i, j;
            const int n = 200;
            int data; int p;
            double K = 1.0 / (double)n;
            double max = -1E10, min = 1E10, v;
            //【AddCount回加算したので平均値を得る】
            for (i = 0; i < ConvPattern.Length; i++)
            {
                for (j = 1; j < n; j++)
                {
                    p = i - j;
                    if (p < 0) { break; }
                    ConvPattern[i] += (double)j * K * IrPattern[p]; //【０シフト＝黒】
                }
                v = ConvPattern[i];
                if (v < min) { min = v; }
                if (v > max) { max = v; }
            }
            v = 65535 / (max - min);
            for (i = j = 0; i < ConvPattern.Length; i++)
            {
                data = (int)((ConvPattern[i] - min) * v);
                if (data < 0) { data = 0; }
                if (data > 65535) { data = 65535; }
                ImageDataXRGB[j] = InfraredColorBar.B[data];
                ImageDataXRGB[j + 1] = InfraredColorBar.G[data];
                ImageDataXRGB[j + 2] = InfraredColorBar.R[data];
                ImageDataXRGB[j + 3] = InfraredColorBar.A[data];
                j += 4;
            }
            bmpdata4 = bmp4.LockBits(rect4, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
            Marshal.Copy(ImageDataXRGB, 0, bmpdata4.Scan0, ImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp4.UnlockBits(bmpdata4);                                              //【Lock解除】
            if (fReverse) { bmp4.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp4;
        }

        //*******************************************************
        //▲【Multiply2】赤外線画像とテンプレートを単純乗算して不一致65535、一致0となる画像を取得する(移動平均なし)
        //*******************************************************
        public void Multiply2()
        {
            int i, j, k;
            double d;
            for (i = j = 0; i < ImageData.Length; i += 2)
            {
                d = (double)(ImageData[i + 1] * 256 + ImageData[i]);
                k = (int)((d - 32768.0) * ConvPattern[j]) + 32768;
                if (k < 0) { k = 0; }
                k = k & 0xffff;
                j++;
                ImageData[i + 1] = (byte)((k >> 8) & 0xff);
                ImageData[i] = (byte)(k & 0xff);
            }
        }


        //*******************************************************
        //【MultiplyPatternsC】Pre-Convoluted Patternと赤外線カメラ画像の畳み込み演算を行う
        //*******************************************************
        public Bitmap ConvolutionPatterns(Boolean fReverse)
        {
            int i, j, k;
            int pX, pXY, pY, nF;
            double d;
            double Mul;
            double cSumX;  //【水平方向累積加算値】
            double cSumXY; //【矩形小領域内累積加算値】

            //【移動平均用FIFOの初期化】
            for (i = 0; i < nSpan; i++) { FIFOx[i] = 0; } //【FIFOxの0クリア】
            for (i = 0; i < nWidth * nSpan; i++) { FIFOxy[i] = 0; }
            for (i = 0; i < nWidth; i++) { SumXY[i] = 0; }
            cSumX = cSumXY = 0;
            pX = pXY = pY = 0;
            nF = nSpan * nWidth;

            //【１フレーム分の赤外線画像データを処理】
            double max = -1E10, min = 1E10, v;
            for (i = j = 0; i < ImageData.Length; i += 2)
            {
                //①【赤外線カメラ画像の１画素データ(輝度)を取得】±32767.0
                d = (double)(ImageData[i + 1] * 256 + ImageData[i]) - 32768.0;

                //②【Pre-Convoluted Code Patternと輝度データを乗算】
                Mul = d * ConvPattern[j];

                //③【FIFOxの移動平均処理】【水平方向】【スキャンライン】
                cSumX += Mul - FIFOx[pX];      //【差分計算から累積加算値を求める】
                FIFOx[pX] = Mul;               //【新規データをFIFOに保存する】
                pX++; if (pX == nSpan) pX = 0; //【FIFOx用ポインタ(0～n-1)の更新処理】

                //④【FIFOxyの移動平均処理】【垂直方向】【スクリーン横幅分並列処理】
                SumXY[pY] += cSumX - FIFOxy[pXY]; //【差分計算から累積加算値を求める】
                FIFOxy[pXY] = cSumX;              //【新規データをFIFOに保存する】
                cSumXY = SumXY[pY];               //【n×n画素の矩形領域の累積加算値⇒add】
                pXY++; if (pXY == nF) pXY = 0;    //【FIFOxy用ポインタ(0～n*width-1)の更新処理】
                pY++; if (pY == nWidth) pY = 0;   //【SumXY用ポインタ(0～width-1)の更新処理】

                //⑤【水平方向シフト量をフレームバッファに書き込む】
                v = cSumXY * nSpan2;
                ConvFrame[j] = v;
                if (v < min) { min = v; }
                if (v > max) { max = v; }
                j++;
            }
            v = 65535.0 / (max - min);
            for (i = j = 0; j < ConvFrame.Length; j++)
            {
                k = 65535 - (int)((ConvFrame[j] - min) * v);
                if (k > 65535) { k = 65535; }
                if (k < 0) { k = 0; }
                ImageDataXRGB[i] = InfraredColorBar.B[k];
                ImageDataXRGB[i + 1] = InfraredColorBar.G[k];
                ImageDataXRGB[i + 2] = InfraredColorBar.R[k];
                ImageDataXRGB[i + 3] = InfraredColorBar.A[k];
                i += 4;
            }
            bmpdata1 = bmp1.LockBits(rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
            Marshal.Copy(ImageDataXRGB, 0, bmpdata1.Scan0, ImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
            bmp1.UnlockBits(bmpdata1);                                              //【Lock解除】
            if (fReverse) { bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
            return bmp1;
        }

        private void CalHosei()
        {
            double K = 1 / 65536.0;
            for (int i = 0; i < 65536; i++)
            {
                Hosei[i] = 65535.0 * Math.Sqrt(Math.Sqrt(i * K));
            }
        }



    }
}
